from enum import Enum
from pathlib import Path

from django.conf import settings
from django.core.files.storage import FileSystemStorage


class TextureType(Enum):
    SKIN = "skin"
    CLOAK = "cloak"
    ELYTRA = "elytra"


class OverwriteStorage(FileSystemStorage):
    def get_available_name(self, name, max_length=None):
        if self.exists(name):
            Path(settings.MEDIA_ROOT, name).unlink()
        return name


def path_handler(texture_type: TextureType, instance, filename: str) -> str:
    return f'{texture_type.value}s/{instance.player.username}.png'



from functools import partial

from django.contrib.auth.models import AbstractUser
from django.contrib.auth.signals import user_logged_in
from django.db import models
from django.dispatch import receiver
from django.utils.timezone import now

from lavender import texture


class Player(AbstractUser):
    packages = models.ManyToManyField('Package', blank=True)


class LogHistory(models.Model):
    player: Player = models.ForeignKey(Player, on_delete=models.CASCADE)

    date = models.DateTimeField(default=now)
    source = models.CharField(max_length=255)

    def __str__(self):
        return f"{self.player.username} at {self.date} from {self.source}"


class Token(models.Model):
    player: Player = models.OneToOneField(Player, on_delete=models.CASCADE)

    access = models.CharField(max_length=32, null=True)
    client = models.CharField(max_length=32, null=True)
    uuid = models.CharField(max_length=32, null=True)
    server_id = models.CharField(max_length=48, null=True)
    created = models.DateTimeField(default=now)

    def __repr__(self):
        return f"{self.player.username} (access: {self.access} client: {self.client}"


class GameLog(models.Model):
    player: Player = models.ForeignKey(Player, on_delete=models.CASCADE)

    date = models.DateTimeField(default=now)
    kind = models.CharField(max_length=16)
    payload = models.TextField(default="")

    def __str__(self):
        return f"{self.player.username} at {self.date}: {self.kind}"


class Package(models.Model):
    name = models.CharField(max_length=16)
    version = models.CharField(max_length=16)
    title = models.CharField(max_length=255)
    location = models.CharField(max_length=255)
    is_default = models.BooleanField(default=False)

    def __str__(self):
        return self.name


class Quenta(models.Model):
    player: Player = models.OneToOneField(Player, on_delete=models.CASCADE, null=True)

    text = models.TextField()
    comments = models.TextField(blank=True)
    approved = models.BooleanField(default=False)

    def __str__(self):
        status = "approved" if self.approved else "declined"
        return f"{self.__class__.__name__} for {self.player.username}: {status}"


class Wardrobe(models.Model):
    player: Player = models.ForeignKey(Player, on_delete=models.CASCADE)

    skin_height = models.PositiveIntegerField(null=True)
    skin_width = models.PositiveIntegerField(null=True)
    skin = models.ImageField(null=True, blank=True, height_field='skin_height', width_field='skin_width',
                             upload_to=partial(texture.path_handler, texture.TextureType.SKIN),
                             storage=texture.OverwriteStorage())

    cloak_height = models.PositiveIntegerField(null=True)
    cloak_width = models.PositiveIntegerField(null=True)
    cloak = models.ImageField(null=True, blank=True, height_field='cloak_height', width_field='cloak_width',
                              upload_to=partial(texture.path_handler, texture.TextureType.CLOAK),
                              storage=texture.OverwriteStorage())

    elytra_height = models.PositiveIntegerField(null=True)
    elytra_width = models.PositiveIntegerField(null=True)
    elytra = models.ImageField(null=True, blank=True, height_field='elytra_height', width_field='elytra_width',
                               upload_to=partial(texture.path_handler, texture.TextureType.ELYTRA),
                               storage=texture.OverwriteStorage())

    def __str__(self):
        return f"{self.__class__.__name__} for {self.player.username}"


@receiver(user_logged_in)
def on_login(sender, user, request, **kwargs):
    record = LogHistory(player=user, source='site')
    record.save()



# Generated by Django 2.2.1 on 2020-01-02 16:57

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('lavender', '0006_blank_packages'),
    ]

    operations = [
        migrations.RenameField(
            model_name='gamelog',
            old_name='log_type',
            new_name='kind',
        ),
        migrations.RemoveField(
            model_name='gamelog',
            name='url',
        ),
        migrations.AddField(
            model_name='gamelog',
            name='payload',
            field=models.TextField(default=''),
        ),
    ]



from functools import wraps
from typing import Callable, Any

from django.http import HttpRequest, HttpResponseForbidden, HttpResponse
from pydantic import BaseModel

from lavender.models import Player

access_header = "X-Lavender-Access"
client_header = "X-Lavender-Client"

Decorator = Callable[[HttpRequest], HttpResponse]


def authenticated(func: Callable[[Player, HttpRequest], BaseModel]):
    @wraps(func)
    def wrapper(request: HttpRequest):
        access_token = request.headers[access_header]
        client_token = request.headers[client_header]
        if not all((access_header, client_header)):
            raise HttpResponseForbidden

        player: Player = Player.objects.get(token__access=access_token, token__client=client_token)
        if player is None:
            raise HttpResponseForbidden

        response = func(player, request)
        return response

    return wrapper


def json_response(func: Callable[[HttpRequest], BaseModel]) -> Decorator:
    @wraps(func)
    def wrapper(request: HttpRequest):
        result = func(request)
        return HttpResponse(result.json(), status=200, content_type='application/json')

    return wrapper



import re

INVALID_CHARS = re.compile(r'[^a-zA-Z_]')
REPEATS = re.compile(r'(?P<char>.+)(?P=char){2,}', re.IGNORECASE)
SIDE_DASHES = re.compile(r'^[_-]|[_-]$')


def check_nickname(nickname: str):
    # В идеале этого не должно происходить, но пусть будет
    if len(nickname) == 0:
        return ['Никнейм не может быть пустым']

    result = []

    if len(nickname) < 3 or len(nickname) > 16:
        result.append('Ник должен быть длиной от 3 до 16 символов')

    if INVALID_CHARS.search(nickname):
        result.append((
            'Ник может состоять только из английских '
            'букв или нижнего подчёркивания'
        ))

    if REPEATS.search(nickname):
        result.append((
            'Ник не должен содержать '
            'повторов из трёх и более символов'
        ))

    if re.search(SIDE_DASHES, nickname):
        result.append((
            'Ник не должен содержать подчёркиваний, '
            'дефисов и прочих украшений по краям'
        ))

    return result



# Generated by Django 2.2.1 on 2019-12-14 18:15

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('lavender', '0005_wardrobe'),
    ]

    operations = [
        migrations.AlterField(
            model_name='player',
            name='packages',
            field=models.ManyToManyField(blank=True, to='lavender.Package'),
        ),
    ]



# Generated by Django 2.2.1 on 2020-01-07 12:11

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('lavender', '0007_fix_gamelog'),
    ]

    operations = [
        migrations.AlterField(
            model_name='token',
            name='player',
            field=models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL),
        ),
    ]



from django import forms
from django.contrib.auth import get_user_model
from django.contrib.auth.forms import UserCreationForm

from lavender.utils import check_nickname
from lavender.models import Quenta, Wardrobe

User = get_user_model()


class RegistrationForm(UserCreationForm):
    email = forms.EmailField(max_length=75, required=True)
    consent = forms.BooleanField(required=True)

    class Meta:
        model = User
        fields = ('username', 'email',)

    def clean_username(self):
        username = self.cleaned_data['username']
        errors = check_nickname(username)
        if len(errors) > 0:
            raise forms.ValidationError(errors)

        return username

    def clean_consent(self):
        pass

    def save(self, commit=True):
        user = super(RegistrationForm, self).save(commit=False)
        user.email = self.cleaned_data["email"]
        if commit:
            user.save()
        return user


class EmailChangeForm(forms.ModelForm):
    email = forms.EmailField(max_length=75, required=False)

    class Meta:
        model = User
        fields = ('email',)

    def clean_email(self):
        email = self.cleaned_data['email']
        if email:
            if User.objects.exclude(username=self.instance.username).filter(email=email).exists():
                raise forms.ValidationError('Этот адрес почты уже используется')
        return email


class QuentaForm(forms.ModelForm):
    class Meta:
        model = Quenta
        fields = ('text',)

    def __init__(self, *args, **kwargs):
        self.user = kwargs.pop("user")
        super().__init__(*args, **kwargs)

    def save(self, commit=True):
        quenta: Quenta = super().save(commit=False)
        quenta.player = self.user
        if commit:
            quenta.save()
        return quenta


class WardrobeForm(forms.ModelForm):
    class Meta:
        model = Wardrobe
        fields = ('skin', 'cloak', 'elytra')

    def __init__(self, *args, **kwargs):
        self.user = kwargs.pop("user")
        super().__init__(*args, **kwargs)

    def save(self, commit=True):
        wardrobe: Wardrobe = super().save(commit=False)
        wardrobe.player = self.user
        if commit:
            wardrobe.save()
        return wardrobe



from importlib import import_module
from typing import Optional

from django.contrib.auth import authenticate
from django.http import HttpResponse, Http404

from lavender import settings
from lavender.models import LogHistory, Player


def _postprocess(user: Optional[Player]) -> bool:
    if user is None:
        return False

    if settings.AUTH_POSTPROCESS_CLASS is None:
        return True

    module_name, class_name = settings.AUTH_POSTPROCESS_CLASS.rsplit(".", 1)
    module = import_module(module_name)
    assert hasattr(module, class_name), f"Class {class_name} is not in {module_name}"
    checker = getattr(module, class_name)()
    return checker.check(user)


def _auth(username, password, handler):
    user = authenticate(username=username, password=password)
    if user:
        record = LogHistory(player=user, source='game')
        record.save()

    return handler(_postprocess(user))


def auth(request, username, password):
    if not settings.AUTH_API_ENABLED:
        raise Http404

    def handler(success: bool) -> HttpResponse:
        status: int = success and 204 or 403
        return HttpResponse(status=status)

    return _auth(username, password, handler)


def auth_legacy(request, username, password):
    if not settings.LEGACY_AUTH_API_ENABLED:
        raise Http404

    def handler(success: bool) -> HttpResponse:
        if success:
            message = f"OK:{username}"
        else:
            message = settings.LEGACY_AUTH_FAIL_MESSAGE
        return HttpResponse(message)

    return _auth(username, password, handler)



from django.conf import settings

AUTH_API_ENABLED = getattr(settings, 'LAVENDER_AUTH_API_ENABLED', False)
LEGACY_AUTH_API_ENABLED = getattr(settings, 'LAVENDER_LEGACY_AUTH_API_ENABLED', False)
LEGACY_AUTH_FAIL_MESSAGE = getattr(settings, 'LAVENDER_LEGACY_AUTH_FAIL_MESSAGE', 'Failed to login')
AUTH_POSTPROCESS_CLASS = getattr(settings, 'LAVENDER_AUTH_POSTPROCESS_CLASS', None)



# Generated by Django 2.2.1 on 2019-06-25 18:21

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('lavender', '0002_add_token_uuid_and_server_id'),
    ]

    operations = [
        migrations.AlterField(
            model_name='token',
            name='server_id',
            field=models.CharField(max_length=48, null=True),
        ),
    ]



# Generated by Django 2.2.1 on 2019-12-08 23:40

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('lavender', '0003_increase_server_id_length'),
    ]

    operations = [
        migrations.CreateModel(
            name='Quenta',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('text', models.TextField()),
                ('comments', models.TextField(blank=True)),
                ('approved', models.BooleanField(default=False)),
                ('player', models.OneToOneField(null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]



# Generated by Django 2.2.1 on 2019-12-09 00:47

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import functools
import lavender.texture


class Migration(migrations.Migration):

    dependencies = [
        ('lavender', '0004_quenta'),
    ]

    operations = [
        migrations.CreateModel(
            name='Wardrobe',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('skin_height', models.PositiveIntegerField(null=True)),
                ('skin_width', models.PositiveIntegerField(null=True)),
                ('skin', models.ImageField(blank=True, height_field='skin_height', null=True, storage=lavender.texture.OverwriteStorage(), upload_to=functools.partial(lavender.texture.path_handler, *(lavender.texture.TextureType('skin'),), **{}), width_field='skin_width')),
                ('cloak_height', models.PositiveIntegerField(null=True)),
                ('cloak_width', models.PositiveIntegerField(null=True)),
                ('cloak', models.ImageField(blank=True, height_field='cloak_height', null=True, storage=lavender.texture.OverwriteStorage(), upload_to=functools.partial(lavender.texture.path_handler, *(lavender.texture.TextureType('cloak'),), **{}), width_field='cloak_width')),
                ('elytra_height', models.PositiveIntegerField(null=True)),
                ('elytra_width', models.PositiveIntegerField(null=True)),
                ('elytra', models.ImageField(blank=True, height_field='elytra_height', null=True, storage=lavender.texture.OverwriteStorage(), upload_to=functools.partial(lavender.texture.path_handler, *(lavender.texture.TextureType('elytra'),), **{}), width_field='elytra_width')),
                ('player', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]



from lavender.forms import EmailChangeForm, QuentaForm, WardrobeForm
from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.contrib.auth.forms import PasswordChangeForm
from django.shortcuts import render

from lavender.models import Quenta, Wardrobe


@login_required
def profile(request):
    email_form = EmailChangeForm(instance=request.user)
    password_form = PasswordChangeForm(user=request.user)
    quenta = Quenta.objects.filter(player=request.user).first()
    wardrobe = Wardrobe.objects.filter(player=request.user).first()
    quenta_form = QuentaForm(user=request.user, instance=quenta)
    wardrobe_form = WardrobeForm(user=request.user, instance=wardrobe)

    if request.method == 'POST':
        if request.POST['form'] == 'email':
            email_form = EmailChangeForm(request.POST, instance=request.user)
            if email_form.is_valid():
                email_form.save()
                messages.success(request, 'Почта успешно изменена!')
            else:
                for field in email_form:
                    for error in field.errors:
                        messages.error(request, error)

        elif request.POST['form'] == 'password':
            password_form = PasswordChangeForm(data=request.POST, user=request.user)
            if password_form.is_valid():
                password_form.save()
                messages.success(request, 'Пароль успешно изменён!')
            else:
                for field in password_form:
                    for error in field.errors:
                        messages.error(request, error)

        elif request.POST['form'] == 'quenta':
            quenta_form = QuentaForm(data=request.POST, user=request.user, instance=quenta)
            if quenta_form.is_valid():
                quenta_form.save()
                messages.success(request, 'Квента успешно сохранена!')
            else:
                for field in quenta_form:
                    for error in field.errors:
                        messages.error(request, error)

        elif request.POST['form'] == 'wardrobe':
            wardrobe_form = WardrobeForm(request.POST, request.FILES, user=request.user, instance=wardrobe)
            if wardrobe_form.is_valid():
                wardrobe_form.save()
                wardrobe = Wardrobe.objects.filter(player=request.user).first()
                messages.success(request, 'Скин успешно сохранен!')
            else:
                for field in quenta_form:
                    for error in field.errors:
                        messages.error(request, error)

    return render(request, 'profile/profile.html', {
        'user': request.user,
        'history': request.user.loghistory_set.order_by('-id')[:10],
        'quenta': quenta,
        'wardrobe': wardrobe,
        'email_form': email_form,
        'password_form': password_form,
        'quenta_form': quenta_form,
        'wardrobe_form': wardrobe_form,
    })



from typing import List

from django.views.decorators.csrf import csrf_exempt
from pydantic import BaseModel

from lavender.models import Package, Player
from lavender.views.api.decorator import authenticated, json_response


class PackageInstance(BaseModel):
    name: str
    title: str
    version: str
    location: str

    class Config:
        allow_mutation = False
        orm_mode = True


class PackageList(BaseModel):
    minimumVersion: int = 3
    packages: List[PackageInstance] = []

    class Config:
        allow_mutation = False


@csrf_exempt
@json_response
@authenticated
def package_list(player: Player, _) -> PackageList:
    packages = [
        PackageInstance.from_orm(package)
        for package in
        Package.objects.filter(player=player)
    ]
    return PackageList(packages=packages)



import json
from typing import List

from django.http import HttpRequest, HttpResponse
from django.views.decorators.csrf import csrf_exempt
from pydantic import BaseModel, parse_obj_as

from lavender.models import Player, GameLog
from lavender.views.api.decorator import authenticated


class LogInstance(BaseModel):
    kind: str
    payload: str


@csrf_exempt
@authenticated
def save_logs(player: Player, request: HttpRequest):
    log_instances: List[LogInstance] = parse_obj_as(List[LogInstance], json.loads(request.body))
    logs = [
        GameLog(player=player, kind=log.kind, payload=log.payload)
        for log
        in log_instances
    ]
    GameLog.objects.bulk_create(logs)
    return HttpResponse()



from lavender.forms import RegistrationForm
from django.contrib.auth import authenticate, login
from django.http import HttpResponseRedirect
from django.shortcuts import redirect, render

from lavender.models import Package


def registration(request):
    if request.user.is_authenticated:
        return HttpResponseRedirect('/accounts/profile')

    if request.method == 'POST':
        form = RegistrationForm(request.POST)
        if form.is_valid():
            form.save()

            username = form.cleaned_data['username']
            raw_password = form.cleaned_data['password1']

            user = authenticate(username=username, password=raw_password)
            user.packages.add(*Package.objects.filter(is_default=True))
            user.save()

            login(request, user)
            return redirect('/accounts/profile')
    else:
        form = RegistrationForm()
    return render(request, 'registration/registration.html', {'form': form})



from django.contrib import admin
from django.contrib.admin import ModelAdmin
from django.contrib.auth.admin import UserAdmin
from django.utils.translation import ugettext_lazy as _

from lavender.models import Package, Player, LogHistory, GameLog, Quenta, Wardrobe


class PlayerAdmin(UserAdmin):
    filter_horizontal = ('packages',)

    def get_fieldsets(self, request, obj=None):
        if not obj:
            return self.add_fieldsets

        if request.user.is_superuser:
            perm_fields = ('is_active', 'is_staff', 'is_superuser',
                           'groups', 'user_permissions')
        else:
            perm_fields = ('is_active', 'is_staff')

        return [
            (None, {'fields': ('username', 'password')}),
            (_('Personal info'), {'fields': ('first_name', 'last_name', 'email')}),
            (_('Permissions'), {'fields': perm_fields}),
            ('Minecraft', {'fields': ('packages',)})
        ]


class LogHistoryAdmin(ModelAdmin):
    list_filter = ('date', 'source')
    list_display = ('player', 'date', 'source')
    search_fields = (
        'player__username',
    )

    def get_queryset(self, request):
        return super().get_queryset(request).select_related('player')


class GameLogAdmin(ModelAdmin):
    list_display = ('player', 'date', 'kind',)
    list_filter = ('date', 'kind')
    search_fields = (
        'player__username',
        'kind'
    )

    def get_queryset(self, request):
        return super().get_queryset(request).select_related('player')


class QuentaAdmin(ModelAdmin):
    list_display = ('player', 'approved',)
    list_filter = ('approved',)
    search_fields = (
        'player__username',
    )

    def get_queryset(self, request):
        return super().get_queryset(request).select_related('player')


class WardrobeAdmin(ModelAdmin):
    list_display = ('player',)
    fields = ('player', 'skin', 'cloak', 'elytra')
    search_fields = (
        'player__username',
    )

    def get_queryset(self, request):
        return super().get_queryset(request).select_related('player')


class PackageAdmin(ModelAdmin):
    list_display = ('name', 'version', 'title', 'location', 'is_default')
    list_filter = ('is_default',)


admin.site.register(Player, PlayerAdmin)
admin.site.register(LogHistory, LogHistoryAdmin)
admin.site.register(GameLog, GameLogAdmin)
admin.site.register(Quenta, QuentaAdmin)
admin.site.register(Wardrobe, WardrobeAdmin)
admin.site.register(Package, PackageAdmin)



import debug_toolbar
from django.conf import settings
from django.conf.urls.static import static
from django.contrib.auth import views as auth
from django.contrib import admin
from django.urls import path, include

from lavender.views import registration, profile
from lavender.views.api import authentication, packages, telemetry

urlpatterns = [
    path('', auth.LoginView.as_view(redirect_authenticated_user=True)),

    path('admin/', admin.site.urls),
    path('accounts/login/', auth.LoginView.as_view(redirect_authenticated_user=True), name='login'),
    path('accounts/logout/', auth.LogoutView.as_view(), name='logout'),

    path('accounts/password_change/', auth.PasswordChangeView.as_view(), name='password_change'),
    path('accounts/password_change/done/', auth.PasswordChangeDoneView.as_view(), name='password_change_done'),

    path('accounts/password_reset/', auth.PasswordResetView.as_view(), name='password_reset'),
    path('accounts/password_reset/done/', auth.PasswordResetDoneView.as_view(), name='password_reset_done'),
    path('accounts/reset/<uidb64>/<token>/', auth.PasswordResetConfirmView.as_view(), name='password_reset_confirm'),
    path('accounts/reset/done/', auth.PasswordResetCompleteView.as_view(), name='password_reset_complete'),

    path('accounts/registration/', registration.registration, name='registration'),
    path('accounts/profile/', profile.profile, name='profile'),

    path('api/v1/authentication/new/<username>/<password>/', authentication.auth),
    path('api/v1/authentication/legacy/<username>/<password>/', authentication.auth_legacy),
    path('api/v1/logs/upload/', telemetry.save_logs),
    path('api/v1/packages/list/', packages.package_list),

    # FIXME: Deprecated and will be removed soon
    path('api/auth/<username>/<password>/', authentication.auth),
    path('api/auth_legacy/<username>/<password>/', authentication.auth_legacy),
    path('api/packages/', packages.package_list),
]

if settings.DEBUG:
    urlpatterns += static(settings.MEDIA_URL, document_root=settings.MEDIA_ROOT)
    urlpatterns += [path('__debug__/', include(debug_toolbar.urls))]
